@@ -15,6 +15,8 @@
  *
  */
 
+#include "config.h"
+
 #ifdef __linux__
 #define _GNU_SOURCE 1		/* strdup() prototype, broken arpa/inet.h */
 #endif
@@ -46,10 +48,17 @@
 #include <resolv.h>
 #include <time.h>
 
-#include "config.h"
+#if defined(__linux__)
+#if defined(HAVE_IN6_ADDR_GEN_MODE_NONE)
+#include <linux/if_link.h>
+#endif // HAVE_IN6_ADDR_GEN_MODE_NONE
+#endif
+
 #include "../lib/tun.h"
 #include "../lib/ippool.h"
 #include "../lib/syserr.h"
+#include "../lib/netns.h"
+#include "../lib/icmpv6.h"
 #include "../gtp/pdp.h"
 #include "../gtp/gtp.h"
 #include "cmdline.h"
@@ -58,15 +67,23 @@
 #define MAXCONTEXTS 1024	/* Max number of allowed contexts */
 
 /* HASH tables for IP address allocation */
+struct pdp_peer_sgsnemu_ctx;
 struct iphash_t {
 	uint8_t inuse;		/* 0=free. 1=used by somebody */
 	struct iphash_t *ipnext;
-	struct pdp_t *pdp;
+	struct pdp_peer_sgsnemu_ctx *ctx;
 	struct in46_addr addr;
 };
-struct iphash_t iparr[MAXCONTEXTS];
 struct iphash_t *iphash[MAXCONTEXTS];
 
+struct pdp_peer_sgsnemu_ctx {
+	struct iphash_t hash_v4;
+	struct iphash_t hash_v6_ll;
+	struct iphash_t hash_v6_global;
+	struct pdp_t *pdp;
+};
+struct pdp_peer_sgsnemu_ctx ctx_arr[MAXCONTEXTS];
+
 /* State variable used for ping  */
 /* 0: Idle                       */
 /* 1: Wait_connect               */
@@ -81,17 +98,21 @@ struct tun_t *tun = NULL;	/* TUN instance */
 int maxfd = 0;			/* For select() */
 int echoversion = 1;		/* First try this version */
 void *tall_sgsnemu_ctx;		/* root talloc ctx */
+#if defined(__linux__)
+int netns = -1;			/* network namespace */
+#endif
 
 /* Struct with local versions of gengetopt options */
 struct {
 	int debug;		/* Print debug messages */
 	int createif;		/* Create local network interface */
 	char *tun_dev_name;
-	struct in46_addr netaddr, destaddr, net;	/* Network interface  */
+	char *netns;
+	struct in46_addr netaddr;	/* Network interface  */
 	size_t prefixlen;
 	char *ipup, *ipdown;	/* Filename of scripts */
 	int defaultroute;	/* Set up default route */
-	struct in_addr pinghost;	/* Remote ping host    */
+	struct in46_addr pinghost;	/* Remote ping host    */
 	int pingrate;
 	int pingsize;
 	int pingcount;
@@ -153,6 +174,11 @@ struct ip_ping {
 	uint8_t data[CREATEPING_MAX];	/* Data */
 } __attribute__ ((packed));
 
+struct ip6_ping {
+	struct icmpv6_echo_hdr hdr;
+	uint8_t data[CREATEPING_MAX];	/* Data */
+} __attribute__ ((packed));
+
 /* Statistical values for ping */
 int nreceived = 0;
 int ntreceived = 0;
@@ -174,6 +200,9 @@ static int ipset(struct iphash_t *ipaddr, struct in46_addr *addr)
 	int hash = ippool_hash(addr) % MAXCONTEXTS;
 	struct iphash_t *h;
 	struct iphash_t *prev = NULL;
+
+	printf("Adding IP to local pool: %s\n", in46a_ntoa(addr));
+
 	ipaddr->ipnext = NULL;
 	ipaddr->addr = *addr;
 	for (h = iphash[hash]; h; h = h->ipnext)
@@ -294,6 +323,8 @@ static int process_options(int argc, char **argv)
 		printf("createif: %d\n", args_info.createif_flag);
 		if (args_info.tun_device_arg)
 			printf("tun-device: %s\n", args_info.tun_device_arg);
+		if (args_info.netns_arg)
+			printf("netns: %s\n", args_info.netns_arg);
 		if (args_info.ipup_arg)
 			printf("ipup: %s\n", args_info.ipup_arg);
 		if (args_info.ipdown_arg)
@@ -352,6 +383,8 @@ static int process_options(int argc, char **argv)
 			printf("createif: %d\n", args_info.createif_flag);
 			if (args_info.tun_device_arg)
 				printf("tun-device: %s\n", args_info.tun_device_arg);
+			if (args_info.netns_arg)
+				printf("netns: %s\n", args_info.netns_arg);
 			if (args_info.ipup_arg)
 				printf("ipup: %s\n", args_info.ipup_arg);
 			if (args_info.ipdown_arg)
@@ -459,27 +492,12 @@ static int process_options(int argc, char **argv)
 	}
 
 	/* imsi                                                            */
-	if (strlen(args_info.imsi_arg) != 15) {
+	if (strlen(args_info.imsi_arg) < 6 || strlen(args_info.imsi_arg) > 15) {
 		printf("Invalid IMSI\n");
 		return -1;
 	}
 
-	options.imsi = 0xf000000000000000ull;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[0] - 48));
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[1] - 48)) << 4;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[2] - 48)) << 8;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[3] - 48)) << 12;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[4] - 48)) << 16;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[5] - 48)) << 20;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[6] - 48)) << 24;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[7] - 48)) << 28;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[8] - 48)) << 32;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[9] - 48)) << 36;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[10] - 48)) << 40;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[11] - 48)) << 44;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[12] - 48)) << 48;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[13] - 48)) << 52;
-	options.imsi |= ((uint64_t) (args_info.imsi_arg[14] - 48)) << 56;
+	options.imsi = gtp_imsi_str2gtp(args_info.imsi_arg);
 
 	printf("IMSI is:               %s (%#08llx)\n",
 	       args_info.imsi_arg, options.imsi);
@@ -870,25 +888,21 @@ static int process_options(int argc, char **argv)
 	/* createif */
 	options.createif = args_info.createif_flag;
 	options.tun_dev_name = args_info.tun_device_arg;
+	options.netns = args_info.netns_arg;
 
 	/* net                                                          */
 	/* Store net as in_addr net and mask                            */
 	if (args_info.net_arg) {
 		if (ippool_aton
-		    (&options.net, &options.prefixlen, args_info.net_arg, 0)) {
+		    (&options.netaddr, &options.prefixlen, args_info.net_arg, 0)) {
 			SYS_ERR(DSGSN, LOGL_ERROR, 0,
 				"Invalid network address: %s!",
 				args_info.net_arg);
 			exit(1);
 		}
-		options.netaddr = options.net;
-		options.destaddr = options.net;
-
 	} else {
-		memset(&options.net, 0, sizeof(options.net));
 		options.prefixlen = 0;
 		memset(&options.netaddr, 0, sizeof(options.netaddr));
-		memset(&options.destaddr, 0, sizeof(options.destaddr));
 	}
 
 	/* ipup */
@@ -903,20 +917,61 @@ static int process_options(int argc, char **argv)
 	/* defaultroute */
 	options.defaultroute = args_info.defaultroute_flag;
 
+	/* PDP Type */
+	if (!strcmp(args_info.pdp_type_arg, "v6"))
+		options.pdp_type = PDP_EUA_TYPE_v6;
+	else if (!strcmp(args_info.pdp_type_arg, "v4"))
+		options.pdp_type = PDP_EUA_TYPE_v4;
+	else {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Unsupported/unknown PDP Type '%s'\n",
+			args_info.pdp_type_arg);
+		return -1;
+	}
+
 	/* pinghost                                                     */
 	/* Store ping host as in_addr                                   */
 	if (args_info.pinghost_arg) {
-		if (!(host = gethostbyname(args_info.pinghost_arg))) {
+		struct addrinfo hints;
+		struct addrinfo *result;
+		memset(&hints, 0, sizeof(struct addrinfo));
+		switch (options.pdp_type) {
+		case PDP_EUA_TYPE_v4:
+			hints.ai_family = AF_INET;
+			break;
+		case PDP_EUA_TYPE_v6:
+			hints.ai_family = AF_INET6;
+			break;
+		default:
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "lookup(AF_UNSPEC) %d", options.pdp_type);
+			hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
+		}
+		hints.ai_socktype = SOCK_DGRAM;
+		hints.ai_flags = 0;
+		hints.ai_protocol = 0;
+		hints.ai_canonname = NULL;
+		hints.ai_addr = NULL;
+		hints.ai_next = NULL;
+		if ((i = getaddrinfo(args_info.pinghost_arg, NULL, &hints, &result)) != 0) {
 			SYS_ERR(DSGSN, LOGL_ERROR, 0,
-				"Invalid ping host: %s!",
-				args_info.pinghost_arg);
+				"Invalid ping host '%s': %s",
+				args_info.pinghost_arg, gai_strerror(i));
 			return -1;
 		} else {
-			memcpy(&options.pinghost.s_addr, host->h_addr,
-			       host->h_length);
+			switch (result->ai_family) {
+			case AF_INET:
+				options.pinghost.len = sizeof(struct in_addr);
+				options.pinghost.v4 = ((struct sockaddr_in*)result->ai_addr)->sin_addr;
+				SYS_ERR(DSGSN, LOGL_ERROR, 0, 	"AF_INET %d", options.pinghost.len);
+				break;
+			case AF_INET6:
+				options.pinghost.len = sizeof(struct in6_addr);
+				options.pinghost.v6 = ((struct sockaddr_in6*)result->ai_addr)->sin6_addr;
+				break;
+			}
 			printf("Using ping host:       %s (%s)\n",
 			       args_info.pinghost_arg,
-			       inet_ntoa(options.pinghost));
+			       in46a_ntoa(&options.pinghost));
+			freeaddrinfo(result);
 		}
 	}
 
@@ -934,59 +989,39 @@ static int process_options(int argc, char **argv)
 	else
 		options.tx_gpdu_seq = 1;
 
-	/* PDP Type */
-	if (!strcmp(args_info.pdp_type_arg, "v6"))
-		options.pdp_type = PDP_EUA_TYPE_v6;
-	else if (!strcmp(args_info.pdp_type_arg, "v4"))
-		options.pdp_type = PDP_EUA_TYPE_v4;
-	else {
-		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Unsupported/unknown PDP Type '%s'\n",
-			args_info.pdp_type_arg);
-		return -1;
-	}
-
-	if (options.pingcount && options.pdp_type != PDP_EUA_TYPE_v4) {
-		SYS_ERR(DSGSN, LOGL_ERROR, 0, "built-in ping only works with IPv4, use tun-device");
-		return -1;
-	}
-
 	return 0;
 
 }
 
-/* read a single value from a /procc file, up to 255 bytes, callee-allocated */
-static char *proc_read(const char *path)
+/* write a single value to a /proc file */
+static int proc_write(const char *path, const char *value)
 {
-	char *ret = NULL;
-	FILE *f;
+        int ret;
+        FILE *f;
 
-	f = fopen(path, "r");
-	if (!f)
-		return NULL;
+        f = fopen(path, "w");
+        if (!f) {
+                SYS_ERR(DSGSN, LOGL_ERROR, 0, "fopen(%s) failed!\n", path);
+                return -1;
+        }
 
-	ret = malloc(256);
-	if (!ret)
-		goto out;
-
-	if (!fgets(ret, 256, f)) {
-		free(ret);
-		ret = NULL;
-		goto out;
-	}
-
-out:
-	fclose(f);
-	return ret;
+        if ((ret = fputs(value, f)) < 0) {
+                SYS_ERR(DSGSN, LOGL_ERROR, 0, "proc_write(%s, %s) failed!\n", path, value);
+        } else {
+                ret = 0;
+        }
+        fclose(f);
+        return ret;
 }
 
-/* Read value of a /proc/sys/net/ipv6/conf file for given device.
+/* Write value of to /proc/sys/net/ipv6/conf file for given device.
  * Memory is dynamically allocated, caller must free it later. */
-static char *proc_ipv6_conf_read(const char *dev, const char *file)
+static int proc_ipv6_conf_write(const char *dev, const char *file, const char *value)
 {
-	const char *fmt = "/proc/sys/net/ipv6/conf/%s/%s";
-	char path[strlen(fmt) + strlen(dev) + strlen(file)+1];
-	snprintf(path, sizeof(path), fmt, dev, file);
-	return proc_read(path);
+        const char *fmt = "/proc/sys/net/ipv6/conf/%s/%s";
+        char path[strlen(fmt) + strlen(dev) + strlen(file)+1];
+        snprintf(path, sizeof(path), fmt, dev, file);
+        return proc_write(path, value);
 }
 
 static char *print_ipprot(int t)
@@ -1109,7 +1144,7 @@ static int ping_timeout(struct timeval *tp)
 	struct timezone tz;
 	struct timeval tv;
 	int diff;
-	if ((options.pinghost.s_addr) && (2 == state) &&
+	if ((options.pinghost.len) && (2 == state) &&
 	    ((pingseq < options.pingcount) || (options.pingcount == 0))) {
 		gettimeofday(&tv, &tz);
 		diff = 1000000 / options.pingrate * pingseq - 1000000 * (tv.tv_sec - firstping.tv_sec) - (tv.tv_usec - firstping.tv_usec);	/* Microseconds safe up to 500 sec */
@@ -1134,7 +1169,7 @@ static int ping_finish()
 	gettimeofday(&tv, &tz);
 	elapsed = 1000000 * (tv.tv_sec - firstping.tv_sec) + (tv.tv_usec - firstping.tv_usec);	/* Microseconds */
 	printf("\n");
-	printf("\n----%s PING Statistics----\n", inet_ntoa(options.pinghost));
+	printf("\n----%s PING Statistics----\n", in46a_ntoa(&options.pinghost));
 	printf("%d packets transmitted in %.3f seconds, ", ntransmitted,
 	       elapsed / 1000000.0);
 	printf("%d packets received, ", nreceived);
@@ -1152,16 +1187,14 @@ static int ping_finish()
 	if (nreceived && tsum)
 		printf("round-trip (ms)  min/avg/max = %.3f/%.3f/%.3f\n\n",
 		       tmin / 1000.0, tsum / 1000.0 / nreceived, tmax / 1000.0);
-	printf("%d packets transmitted \n", ntreceived);
+	printf("%d packets transmitted \n", ntransmitted);
 
 	ntransmitted = 0;
 	return 0;
 }
 
-/* Handle a received ping packet. Print out line and update statistics. */
-static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
+static int encaps_ping4(struct pdp_t *pdp, void *pack, unsigned len)
 {
-	struct timezone tz;
 	struct timeval tv;
 	struct timeval *tp;
 	struct ip_ping *pingpack = pack;
@@ -1170,17 +1203,12 @@ static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
 
 	src.s_addr = pingpack->src;
 
-	gettimeofday(&tv, &tz);
-	if (options.debug)
-		printf("%d.%6d ", (int)tv.tv_sec, (int)tv.tv_usec);
-
 	if (len < CREATEPING_IP + CREATEPING_ICMP) {
 		printf("packet too short (%d bytes) from %s\n", len,
 		       inet_ntoa(src));
 		return 0;
 	}
 
-	ntreceived++;
 	if (pingpack->protocol != 1) {
 		if (!options.pingquiet)
 			printf("%d bytes from %s: ip_protocol=%d (%s)\n",
@@ -1204,7 +1232,7 @@ static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
 		       inet_ntoa(src), ntohs(pingpack->seq));
 
 	if (len >= sizeof(struct timeval) + CREATEPING_IP + CREATEPING_ICMP) {
-		gettimeofday(&tv, &tz);
+		gettimeofday(&tv, NULL);
 		tp = (struct timeval *)pingpack->data;
 		if ((tv.tv_usec -= tp->tv_usec) < 0) {
 			tv.tv_sec--;
@@ -1227,15 +1255,106 @@ static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
 	return 0;
 }
 
-/* Create a new ping packet and send it off to peer. */
-static int create_ping(void *gsn, struct pdp_t *pdp,
-			struct in_addr *dst, int seq, unsigned int datasize)
+static int encaps_ping6(struct pdp_t *pdp, struct ip6_hdr *ip6h, unsigned len)
 {
+	const struct icmpv6_echo_hdr *ic6h = (struct icmpv6_echo_hdr *) ((uint8_t*)ip6h + sizeof(*ip6h));
+	struct timeval tv;
+	struct timeval tp;
+	int triptime;
+	char straddr[128];
 
+	if (len < sizeof(struct ip6_hdr)) {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Packet len too small to contain IPv6 header (%d)", len);
+		return 0;
+	}
+
+	if (ip6h->ip6_ctlun.ip6_un1.ip6_un1_nxt != IPPROTO_ICMPV6) {
+		if (!options.pingquiet)
+			printf("%d bytes from %s: ip6_protocol=%d (%s)\n", len,
+			       inet_ntop(AF_INET6, &ip6h->ip6_src, straddr, sizeof(straddr)),
+			       ip6h->ip6_ctlun.ip6_un1.ip6_un1_nxt,
+			       print_ipprot(ip6h->ip6_ctlun.ip6_un1.ip6_un1_nxt));
+		return 0;
+	}
+
+	if (len < sizeof(struct ip6_hdr) + sizeof(struct icmpv6_echo_hdr)) {
+		LOGP(DSGSN, LOGL_ERROR, "Packet len too small to contain ICMPv6 echo header (%d)\n", len);
+		return 0;
+	}
+
+	if (ic6h->hdr.type != 129 || ic6h->hdr.code != 0) {
+		if (!options.pingquiet)
+			printf
+			    ("%d bytes from %s: icmp_type=%d icmp_code=%d\n", len,
+			    inet_ntop(AF_INET6, &ip6h->ip6_src, straddr, sizeof(straddr)),
+			    ic6h->hdr.type, ic6h->hdr.code);
+		return 0;
+	}
+
+	nreceived++;
+	if (!options.pingquiet)
+		printf("%d bytes from %s: icmp_seq=%d", len,
+		       inet_ntop(AF_INET6, &ip6h->ip6_src, straddr, sizeof(straddr)),
+		       ntohs(ic6h->seq));
+
+	if (len >= sizeof(struct ip6_hdr) + sizeof(struct icmpv6_echo_hdr) + sizeof(struct timeval)) {
+		gettimeofday(&tv, NULL);
+		memcpy(&tp, ic6h->data, sizeof(struct timeval));
+		if ((tv.tv_usec -= tp.tv_usec) < 0) {
+			tv.tv_sec--;
+			tv.tv_usec += 1000000;
+		}
+		tv.tv_sec -= tp.tv_sec;
+
+		triptime = tv.tv_sec * 1000000 + (tv.tv_usec);
+		tsum += triptime;
+		if (triptime < tmin)
+			tmin = triptime;
+		if (triptime > tmax)
+			tmax = triptime;
+
+		if (!options.pingquiet)
+			printf(" time=%.3f ms\n", triptime / 1000.0);
+
+	} else if (!options.pingquiet)
+		printf("\n");
+	return 0;
+}
+
+/* Handle a received ping packet. Print out line and update statistics. */
+static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
+{
+	struct iphdr *iph = (struct iphdr *)pack;
+	struct timeval tv;
+
+
+	gettimeofday(&tv, NULL);
+	if (options.debug)
+		printf("%d.%6d ", (int)tv.tv_sec, (int)tv.tv_usec);
+
+	ntreceived++;
+
+	if (len < sizeof(struct iphdr)) {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Packet len too small to contain ip header (%d)", len);
+		return -1;
+	}
+	switch(iph->version) {
+	case 4:
+		return encaps_ping4(pdp, pack, len);
+	case 6:
+		return encaps_ping6(pdp, (struct ip6_hdr *)pack, len);
+	default:
+		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Unknown ip header version %d", iph->version);
+		return -1;
+	}
+}
+
+static int create_ping4(void *gsn, struct pdp_t *pdp, struct in46_addr *src,
+			struct in46_addr *dst, int seq, unsigned int datasize)
+{
 	struct ip_ping pack;
 	uint16_t v16;
 	uint8_t *p8 = (uint8_t *) & pack;
-	struct in_addr src;
 	unsigned int n;
 	long int sum = 0;
 	int count = 0;
@@ -1244,14 +1363,6 @@ static int create_ping(void *gsn, struct pdp_t *pdp,
 	struct timeval *tp =
 	    (struct timeval *)&p8[CREATEPING_IP + CREATEPING_ICMP];
 
-	if (datasize > CREATEPING_MAX) {
-		SYS_ERR(DSGSN, LOGL_ERROR, 0,
-			"Ping size to large: %d!", datasize);
-		return -1;
-	}
-
-	memcpy(&src, &(pdp->eua.v[2]), 4);	/* Copy a 4 byte address */
-
 	pack.ipver = 0x45;
 	pack.tos = 0x00;
 	pack.length = htons(CREATEPING_IP + CREATEPING_ICMP + datasize);
@@ -1260,8 +1371,8 @@ static int create_ping(void *gsn, struct pdp_t *pdp,
 	pack.ttl = 0x40;
 	pack.protocol = 0x01;
 	pack.ipcheck = 0x0000;
-	pack.src = src.s_addr;
-	pack.dst = dst->s_addr;
+	pack.src = src->v4.s_addr;
+	pack.dst = dst->v4.s_addr;
 	pack.type = 0x08;
 	pack.code = 0x00;
 	pack.checksum = 0x0000;
@@ -1311,12 +1422,102 @@ static int create_ping(void *gsn, struct pdp_t *pdp,
 	return gtp_data_req(gsn, pdp, &pack, 28 + datasize);
 }
 
+static int create_ping6(void *gsn, struct pdp_t *pdp, struct in46_addr *src,
+			struct in46_addr *dst, int seq, unsigned int datasize)
+{
+	struct ip6_ping *pack;
+	uint8_t *p8;
+	unsigned int n;
+	struct timezone tz;
+	struct timeval *tp;
+
+	struct msgb *msg = msgb_alloc_headroom(sizeof(struct ip6_ping) + 128,128, "ICMPv6 echo");
+	OSMO_ASSERT(msg);
+	pack = (struct ip6_ping *) msgb_put(msg, sizeof(struct icmpv6_echo_hdr) + datasize);
+	pack->hdr.hdr.type = 128;
+	pack->hdr.hdr.code = 0;
+	pack->hdr.hdr.csum = 0;  /* updated below */
+	pack->hdr.ident = 0x0000;
+	pack->hdr.seq = htons(seq);
+
+	p8 = pack->data;
+	for (n = 0; n < (datasize); n++)
+		p8[n] = n;
+
+	if (datasize >= sizeof(struct timeval)) {
+		tp = (struct timeval *)pack->data;
+		gettimeofday(tp, &tz);
+	}
+
+	pack->hdr.hdr.csum = icmpv6_prepend_ip6hdr(msg, &src->v6, &dst->v6);
+
+	ntransmitted++;
+	return gtp_data_req(gsn, pdp, msgb_data(msg), msgb_length(msg));
+}
+
+/* Create a new ping packet and send it off to peer. */
+static int create_ping(void *gsn, struct pdp_t *pdp,
+			struct in46_addr *dst, int seq, unsigned int datasize)
+{
+	int num_addr;
+	struct in46_addr addr[2];
+	struct in46_addr *src;
+
+	if (datasize > CREATEPING_MAX) {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0,
+			"Ping size to large: %d!", datasize);
+		return -1;
+	}
+
+	if ((num_addr = in46a_from_eua(&pdp->eua, addr)) < 1) {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0,
+			"in46a_from_eua() failed! %d", num_addr);
+		return -1;
+	}
+	if (dst->len == addr[0].len) {
+		src = &addr[0];
+	} else if (num_addr > 1 && dst->len == addr[1].len) {
+		src = &addr[1];
+	} else {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0,
+			"Mismaching source and destination IP addr types (%d vs %d)", dst->len, addr[0].len);
+		return -1;
+	}
+	if (in46a_is_v4(dst))
+		return create_ping4(gsn, pdp, src, dst, seq, datasize);
+	else
+		return create_ping6(gsn, pdp, src, dst, seq, datasize);
+}
+
 static int delete_context(struct pdp_t *pdp)
 {
+	int rc;
 
-	if (tun && options.ipdown)
+	if (tun && options.ipdown) {
+#if defined(__linux__)
+		sigset_t oldmask;
+
+		if ((options.netns)) {
+			if ((rc = switch_ns(netns, &oldmask)) < 0) {
+				SYS_ERR(DSGSN, LOGL_ERROR, 0,
+					"Failed to switch to netns %s: %s\n",
+					options.netns, strerror(-rc));
+			}
+		}
+#endif
 		tun_runscript(tun, options.ipdown);
 
+#if defined(__linux__)
+		if ((options.netns)) {
+			if ((rc = restore_ns(&oldmask)) < 0) {
+				SYS_ERR(DSGSN, LOGL_ERROR, 0,
+					"Failed to switch to original netns: %s\n",
+					strerror(-rc));
+			}
+		}
+#endif
+	}
+
 	ipdel((struct iphash_t *)pdp->peer[0]);
 	memset(pdp->peer[0], 0, sizeof(struct iphash_t));	/* To be sure */
 
@@ -1345,19 +1546,8 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 		src.len = 4;
 		src.v4.s_addr = iph->saddr;
 	} else if (iph->version == 6) {
-		/* We only have a single entry in the hash table, and it consists of the link-local
-		 * address "fe80::prefix".  So we need to make sure to convert non-link-local source
-		 * addresses to that format before looking up the hash table via ippool_getip() */
 		src.len = 16;
-		if (!memcmp(ip6h->ip6_src.s6_addr, ll_prefix, sizeof(ll_prefix))) {
-			/* is a link-local address, we can do the hash lookup 1:1 */
-			src.v6 = ip6h->ip6_src;
-		} else {
-			/* it is not a link-local address, so we must convert from the /64 prefix
-			 * to the link-local format that's used in the hash table */
-			memcpy(&src.v6.s6_addr[0], ll_prefix, sizeof(ll_prefix));
-			memcpy(&src.v6.s6_addr[sizeof(ll_prefix)], ip6h->ip6_src.s6_addr, 16-sizeof(ll_prefix));
-		}
+		src.v6 = ip6h->ip6_src;
 	} else {
 		printf("Dropping packet with invalid IP version %u\n", iph->version);
 		return 0;
@@ -1369,28 +1559,32 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 		return 0;
 	}
 
-	if (ipm->pdp)		/* Check if a peer protocol is defined */
-		gtp_data_req(gsn, ipm->pdp, pack, len);
+	if (ipm->ctx->pdp)		/* Check if a peer protocol is defined */
+		gtp_data_req(gsn, ipm->ctx->pdp, pack, len);
 	return 0;
 }
 
 static int create_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
 {
-	struct in46_addr addr;
+	int rc, i, num_addr;
+	struct in46_addr addr[2];
+#if defined(__linux__)
+	sigset_t oldmask;
+#endif
 
-	struct iphash_t *iph = (struct iphash_t *)cbp;
+	struct pdp_peer_sgsnemu_ctx *ctx = (struct pdp_peer_sgsnemu_ctx *) cbp;
 
 	if (cause < 0) {
 		printf("Create PDP Context Request timed out\n");
-		if (iph->pdp->version == 1) {
+		if (ctx->pdp->version == 1) {
 			printf("Retrying with version 0\n");
-			iph->pdp->version = 0;
-			gtp_create_context_req(gsn, iph->pdp, iph);
+			ctx->pdp->version = 0;
+			gtp_create_context_req(gsn, ctx->pdp, ctx);
 			return 0;
 		} else {
 			state = 0;
-			pdp_freepdp(iph->pdp);
-			iph->pdp = NULL;
+			pdp_freepdp(ctx->pdp);
+			ctx->pdp = NULL;
 			return EOF;
 		}
 	}
@@ -1400,77 +1594,99 @@ static int create_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
 		    ("Received create PDP context response. Cause value: %d\n",
 		     cause);
 		state = 0;
-		pdp_freepdp(iph->pdp);
-		iph->pdp = NULL;
+		pdp_freepdp(ctx->pdp);
+		ctx->pdp = NULL;
 		return EOF;	/* Not what we expected */
 	}
 
-	if (in46a_from_eua(&pdp->eua, &addr) < 1) {
+	if ((num_addr = in46a_from_eua(&pdp->eua, addr)) < 1) {
 		printf
 		    ("Received create PDP context response. Cause value: %d\n",
 		     cause);
-		pdp_freepdp(iph->pdp);
-		iph->pdp = NULL;
+		pdp_freepdp(ctx->pdp);
+		ctx->pdp = NULL;
 		state = 0;
 		return EOF;	/* Not a valid IP address */
 	}
 
-	printf("Received create PDP context response. IP address: %s\n",
-	       in46a_ntoa(&addr));
+	printf("Received create PDP context response.\n");
 
-	switch (addr.len) {
-	case 16: /* IPv6 */
-		/* we have to enable the kernel to perform stateless autoconfiguration,
-		 * i.e. send a router solicitation using the lover 64bits of the allocated
-		 * EUA as interface identifier, as per 3GPP TS 29.061 Section 11.2.1.3.2 */
-		memcpy(addr.v6.s6_addr, ll_prefix, sizeof(ll_prefix));
-		printf("Derived IPv6 link-local address: %s\n", in46a_ntoa(&addr));
-		break;
-	case 4: /* IPv4 */
-		break;
+#if defined(__linux__)
+	if ((options.createif) && (options.netns)) {
+		if ((rc = switch_ns(netns, &oldmask)) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0,
+				"Failed to switch to netns %s: %s\n",
+				options.netns, strerror(-rc));
+		}
 	}
+#endif
 
-	if ((options.createif) && (!options.net.len)) {
-		size_t prefixlen = 32;
-		if (addr.len == 16)
-			prefixlen = 64;
-		/* printf("Setting up interface and routing\n"); */
-		tun_addaddr(tun, &addr, &addr, prefixlen);
-		if (options.defaultroute) {
-			struct in_addr rm;
-			rm.s_addr = 0;
-			netdev_addroute(&rm, &addr.v4, &rm);
+	for (i = 0; i < num_addr; i++) {
+		printf("PDP ctx: received EUA with IP address: %s\n",  in46a_ntoa(&addr[i]));
+
+		switch (addr[i].len) {
+		case 16: /* IPv6 */
+			/* Convert address to link local using the lower 64bits
+			   of the allocated EUA as Interface-Identifier to
+			   send router solicitation, as per 3GPP TS 29.061
+			   Section 11.2.1.3.2 */
+			memcpy(addr[i].v6.s6_addr, ll_prefix, sizeof(ll_prefix));
+			printf("Derived IPv6 link-local address: %s\n", in46a_ntoa(&addr[i]));
+			ctx->hash_v6_ll.inuse = 1;
+			ipset(&ctx->hash_v6_ll, &addr[i]);
+			break;
+		case 4: /* IPv4 */
+			ctx->hash_v4.inuse = 1;
+			ipset(&ctx->hash_v4, &addr[i]);
+			break;
+		}
+
+		if ((options.createif) && (!options.netaddr.len)) {
+			size_t prefixlen = 32;
+			if (addr[i].len == 16)
+				prefixlen = 64;
+			/* printf("Setting up interface and routing\n"); */
+			tun_addaddr(tun, &addr[i], NULL, prefixlen);
+			if (options.defaultroute) {
+				if (in46a_is_v4(&addr[i])) {
+					struct in_addr rm;
+					rm.s_addr = 0;
+					if (netdev_addroute4(&rm, &addr[i].v4, &rm) < 0) {
+						SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed adding default route to %s", in46a_ntoa(&addr[i]));
+					}
+				} /* else: route will be set up once we have a global link address (Router Advertisement) */
+			}
+			if (options.ipup)
+				tun_runscript(tun, options.ipup);
 		}
-		if (options.ipup)
-			tun_runscript(tun, options.ipup);
 	}
 
-	/* now that ip-up has been executed, check if we are configured to
-	 * accept router advertisements */
 	if (options.createif && options.pdp_type == PDP_EUA_TYPE_v6) {
-		char *accept_ra, *forwarding;
-
-		accept_ra = proc_ipv6_conf_read(tun->devname, "accept_ra");
-		forwarding = proc_ipv6_conf_read(tun->devname, "forwarding");
-		if (!accept_ra || !forwarding)
-			printf("Could not open proc file for %s ?!?\n", tun->devname);
-		else {
-			if (!strcmp(accept_ra, "0")) {
-				printf("accept_ra=0, i.e. your tun device is not configured to accept "
-					"router advertisements; SLAAC will not succeed, please "
-					"fix your setup!\n");
-			}
-			if (!strcmp(forwarding, "1") && !strcmp(accept_ra, "1")) {
-				printf("forwarding=1 and accept_ra=1, i.e. your tun device is not "
-					"configured to accept router advertisements; SLAAC will not "
-					"succeed, please fix your setup!\n");
-			}
+		struct in6_addr *saddr6;
+		struct msgb *msg;
+		if (in46a_is_v6(&addr[0])) {
+			saddr6 = &addr[0].v6;
+		} else if (num_addr > 1 && in46a_is_v6(&addr[1])) {
+			saddr6 = &addr[1].v6;
+		} else {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to find IPv6 EUA on IPv6 APN");
+			return EOF;	/* Not a valid IP address */
 		}
-		free(accept_ra);
-		free(forwarding);
+		SYS_ERR(DSGSN, LOGL_INFO, 0, "Sending ICMPv6 Router Soliciation to GGSN...");
+		msg = icmpv6_construct_rs(saddr6);
+		gtp_data_req(gsn, ctx->pdp, msgb_data(msg), msgb_length(msg));
+		msgb_free(msg);
 	}
 
-	ipset(iph, &addr);
+#if defined(__linux__)
+	if ((options.createif) && (options.netns)) {
+		if ((rc = restore_ns(&oldmask)) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to switch to original netns: %s\n",
+				strerror(-rc));
+		}
+	}
+#endif
+
 
 	state = 2;		/* Connected */
 
@@ -1508,7 +1724,7 @@ static int echo_conf(int recovery)
 	return 0;
 }
 
-static int conf(int type, int cause, struct pdp_t *pdp, void *cbp)
+static int _gtp_cb_conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 {
 	/* if (cause < 0) return 0; Some error occurred. We don't care */
 	switch (type) {
@@ -1523,8 +1739,90 @@ static int conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 	}
 }
 
+static void handle_router_adv(struct pdp_t *pdp, struct ip6_hdr *ip6h, struct icmpv6_radv_hdr *ra, size_t ra_len)
+{
+	struct pdp_peer_sgsnemu_ctx* ctx = (struct pdp_peer_sgsnemu_ctx*)pdp->peer[0];
+	struct icmpv6_opt_hdr *opt_hdr;
+	struct icmpv6_opt_prefix *opt_prefix;
+	int rc;
+	sigset_t oldmask;
+	struct in6_addr rm;
+	char ip6strbuf[200];
+	memset(&rm, 0, sizeof(rm));
+
+	SYS_ERR(DSGSN, LOGL_INFO, 0, "Received ICMPv6 Router Advertisement");
+
+	foreach_icmpv6_opt(ra, ra_len, opt_hdr) {
+		if (opt_hdr->type == ICMPv6_OPT_TYPE_PREFIX_INFO) {
+			opt_prefix = (struct icmpv6_opt_prefix *)opt_hdr;
+			size_t prefix_len_bytes = (opt_prefix->prefix_len + 7)/8;
+			SYS_ERR(DSGSN, LOGL_INFO, 0, "Parsing OPT Prefix info (prefix_len=%u): %s",
+				opt_prefix->prefix_len,
+				osmo_hexdump((const unsigned char *)opt_prefix->prefix, prefix_len_bytes));
+			if ((options.createif) && (!options.netaddr.len)) {
+				struct in46_addr addr;
+				addr.len = 16;
+				memcpy(addr.v6.s6_addr, opt_prefix->prefix, prefix_len_bytes);
+				memset(&addr.v6.s6_addr[prefix_len_bytes], 0, 16 - prefix_len_bytes);
+				addr.v6.s6_addr[15] = 0x02;
+				SYS_ERR(DSGSN, LOGL_INFO, 0, "Adding addr %s to tun %s",
+					in46a_ntoa(&addr), tun->devname);
+				if (!ctx->hash_v6_global.inuse) {
+					ctx->hash_v6_global.inuse = 1;
+					ipset(&ctx->hash_v6_global, &addr);
+				} else {
+					SYS_ERR(DSGSN, LOGL_ERROR, 0, "First v6 global address in hash already in use!");
+				}
+
+#if defined(__linux__)
+				if ((options.netns)) {
+					if ((rc = switch_ns(netns, &oldmask)) < 0) {
+						SYS_ERR(DSGSN, LOGL_ERROR, 0,
+							"Failed to switch to netns %s: %s",
+							options.netns, strerror(-rc));
+					}
+				}
+#endif
+				rc = tun_addaddr(tun, &addr, NULL, opt_prefix->prefix_len);
+				if (rc < 0) {
+					SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to add addr %s to tun %s",
+						in46a_ntoa(&addr), tun->devname);
+				}
+
+				struct in6_addr rm;
+				memset(&rm, 0, sizeof(rm));
+				if (netdev_addroute6(&rm, &ip6h->ip6_src, 0, tun->devname) < 0) {
+					SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed adding default route to %s", inet_ntop(AF_INET6, &ip6h->ip6_src, ip6strbuf, sizeof(ip6strbuf)));
+				}
+
+#if defined(__linux__)
+				if ((options.netns)) {
+					if ((rc = restore_ns(&oldmask)) < 0) {
+						SYS_ERR(DSGSN, LOGL_ERROR, 0,
+							"Failed to switch to original netns: %s",
+							strerror(-rc));
+					}
+				}
+#endif
+			}
+		}
+	}
+}
+
 static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 {
+	struct iphdr *iph = (struct iphdr *)pack;
+	struct icmpv6_radv_hdr *ra;
+	switch (iph->version) {
+	case 6:
+		if ((ra = icmpv6_validate_router_adv(pack, len))) {
+			size_t ra_len = (uint8_t*)ra - (uint8_t*)pack;
+			handle_router_adv(pdp, (struct ip6_hdr *)pack, ra, ra_len);
+			return 0;
+		}
+	break;
+	}
+
 	/*  printf("encaps_tun. Packet received: forwarding to tun\n"); */
 	return tun_encaps((struct tun_t *)pdp->ipif, pack, len);
 }
@@ -1534,7 +1832,7 @@ int main(int argc, char **argv)
 	fd_set fds;		/* For select() */
 	struct timeval idleTime;	/* How long to select() */
 	struct pdp_t *pdp;
-	int n;
+	int n, rc;
 	int starttime = time(NULL);	/* Time program was started */
 	int stoptime = 0;	/* Time to exit */
 	int pingtimeout = 0;	/* Time to print ping statistics */
@@ -1543,6 +1841,10 @@ int main(int argc, char **argv)
 	struct timezone tz;	/* Used for calculating ping times */
 	struct timeval tv;
 	int diff;
+#if defined(__linux__)
+	char buf[10];
+	sigset_t oldmask;
+#endif
 
 	signal(SIGTERM, signal_handler);
 	signal(SIGHUP,  signal_handler);
@@ -1552,6 +1854,13 @@ int main(int argc, char **argv)
 	msgb_talloc_ctx_init(tall_sgsnemu_ctx, 0);
 	osmo_init_logging2(tall_sgsnemu_ctx, &log_info);
 
+#if defined(__linux__)
+	if ((rc = init_netns()) < 0) {
+		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to initialize netns: %s", strerror(-rc));
+		exit(1);
+	}
+#endif
+
 	/* Process options given in configuration file and command line */
 	if (process_options(argc, argv))
 		exit(1);
@@ -1569,12 +1878,27 @@ int main(int argc, char **argv)
 		maxfd = gsn->fd1u;
 
 	gtp_set_cb_delete_context(gsn, delete_context);
-	gtp_set_cb_conf(gsn, conf);
+	gtp_set_cb_conf(gsn, _gtp_cb_conf);
 	if (options.createif)
 		gtp_set_cb_data_ind(gsn, encaps_tun);
 	else
 		gtp_set_cb_data_ind(gsn, encaps_ping);
 
+#if defined(__linux__)
+	if ((options.createif) && (options.netns)) {
+		if ((netns = get_nsfd(options.netns)) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to obtain fd for netns %s: %s\n",
+				options.netns, strerror(-netns));
+			exit(1);
+		}
+		if ((rc = switch_ns(netns, &oldmask)) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to switch to netns %s: %s\n",
+				options.netns, strerror(-rc));
+			exit(1);
+		}
+	}
+#endif
+
 	if (options.createif) {
 		printf("Setting up interface\n");
 		/* Create a tunnel interface */
@@ -1583,26 +1907,61 @@ int main(int argc, char **argv)
 				"Failed to create tun");
 			exit(1);
 		}
+
+#if defined(__linux__) && defined(HAVE_IN6_ADDR_GEN_MODE_NONE)
+		/* Avoid tunnel setting its own link-local addr automatically,
+		   we don't need it. Don't exit on error since this sysctl is
+		   only available starting with linux 4.11. */
+		snprintf(buf, sizeof(buf), "%u", IN6_ADDR_GEN_MODE_NONE);
+		if (proc_ipv6_conf_write(tun->devname, "addr_gen_mode", buf) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, errno,
+				"Failed to disable addr_gen_mode on %s, an extra link-local "
+				"ip address will appear on the tun device.\n",
+				tun->devname);
+		}
+#endif
+
 		tun_set_cb_ind(tun, cb_tun_ind);
 		if (tun->fd > maxfd)
 			maxfd = tun->fd;
+
+		if (proc_ipv6_conf_write(tun->devname, "accept_ra", "0") < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0,
+				"Failed to disable IPv6 SLAAC on %s\n", tun->devname);
+			exit(1);
+		}
 	}
 
-	if ((options.createif) && (options.net.len)) {
-		/* printf("Setting up interface and routing\n"); */
-		tun_addaddr(tun, &options.netaddr, &options.destaddr, options.prefixlen);
+	if ((options.createif) && (options.netaddr.len)) {
+		tun_addaddr(tun, &options.netaddr, NULL, options.prefixlen);
 		if (options.defaultroute) {
-			struct in_addr rm;
-			rm.s_addr = 0;
-			netdev_addroute(&rm, &options.destaddr.v4, &rm);
+			if (in46a_is_v4(&options.netaddr)) {
+				struct in_addr rm;
+				rm.s_addr = 0;
+				netdev_addroute4(&rm, &options.netaddr.v4, &rm);
+			} else {
+				struct in6_addr rm;
+				memset(&rm, 0, sizeof(rm));
+				netdev_addroute6(&rm, &options.netaddr.v6, 0, tun->devname);
+			}
 		}
 		if (options.ipup)
 			tun_runscript(tun, options.ipup);
 	}
 
+#if defined(__linux__)
+	if ((options.createif) && (options.netns)) {
+		if ((rc = restore_ns(&oldmask)) < 0) {
+			SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to switch to original netns: %s\n",
+				strerror(-rc));
+			exit(1);
+		}
+	}
+#endif
+
 	/* Initialise hash tables */
 	memset(&iphash, 0, sizeof(iphash));
-	memset(&iparr, 0, sizeof(iparr));
+	memset(&ctx_arr, 0, sizeof(ctx_arr));
 
 	printf("Done initialising GTP library\n\n");
 
@@ -1614,7 +1973,6 @@ int main(int argc, char **argv)
 	for (n = 0; n < options.contexts; n++) {
 		uint64_t myimsi;
 		printf("Setting up PDP context #%d\n", n);
-		iparr[n].inuse = 1;	/* TODO */
 
 		imsi_add(options.imsi, &myimsi, n);
 
@@ -1623,9 +1981,12 @@ int main(int argc, char **argv)
 		/* Otherwise it is deallocated by gtplib */
 		gtp_pdp_newpdp(gsn, &pdp, myimsi, options.nsapi, NULL);
 
-		pdp->peer[0] = &iparr[n]; /* FIXME: support v4v6, have 2 peers */
+		pdp->peer[0] = &ctx_arr[n];
 		pdp->ipif = tun;	/* TODO */
-		iparr[n].pdp = pdp;
+		ctx_arr[n].pdp = pdp;
+		ctx_arr[n].hash_v4.ctx = &ctx_arr[n];
+		ctx_arr[n].hash_v6_ll.ctx = &ctx_arr[n];
+		ctx_arr[n].hash_v6_global.ctx = &ctx_arr[n];
 
 		if (options.gtpversion == 0) {
 			if (options.qos.l - 1 > sizeof(pdp->qos_req0)) {
@@ -1713,7 +2074,7 @@ int main(int argc, char **argv)
 
 		/* Create context */
 		/* We send this of once. Retransmissions are handled by gtplib */
-		gtp_create_context_req(gsn, pdp, &iparr[n]);
+		gtp_create_context_req(gsn, pdp, &ctx_arr[n]);
 	}
 
 	state = 1;		/* Enter wait_connection state */
@@ -1763,8 +2124,8 @@ int main(int argc, char **argv)
 			for (n = 0; n < options.contexts; n++) {
 				/* Delete context */
 				printf("Disconnecting PDP context #%d\n", n);
-				gtp_delete_context_req2(gsn, iparr[n].pdp, NULL, 1);
-				if ((options.pinghost.s_addr != 0)
+				gtp_delete_context_req2(gsn, ctx_arr[n].pdp, NULL, 1);
+				if ((options.pinghost.len)
 				    && ntransmitted)
 					ping_finish();
 			}
@@ -1774,7 +2135,7 @@ int main(int argc, char **argv)
 		diff = 0;
 		while ((diff <= 0) &&
 		       /* Send off an ICMP ping packet */
-		       /*if ( */ (options.pinghost.s_addr) && (2 == state) &&
+		       /*if ( */ (options.pinghost.len) && (2 == state) &&
 		       ((pingseq < options.pingcount)
 			|| (options.pingcount == 0))) {
 			if (!pingseq)
@@ -1785,7 +2146,7 @@ int main(int argc, char **argv)
 				if (options.debug)
 					printf("Create_ping %d\n", diff);
 				create_ping(gsn,
-					    iparr[pingseq %
+					    ctx_arr[pingseq %
 						  options.contexts].pdp,
 					    &options.pinghost, pingseq,
 					    options.pingsize);
